#!/usr/bin/fades

import os
import re
import subprocess
import sys
import tempfile
from contextlib import contextmanager

from moviepy.editor import VideoFileClip  # fades

import subtitles  # fades subliminal

SUB_LOADERS = {
    'srt': subtitles.load_srt,
    'ssa': subtitles.load_ssa,
}

RE_TRACK = re.compile(r"Track ID (?P<trackid>\d+): (?P<kind>\w+) \((?P<spec>.+)\).*")


@contextmanager
def lang_changer(new_lang):
    """Simple context manager to change the LANG in the environment."""
    old_lang = os.environ['LANG']
    os.environ['LANG'] = new_lang
    yield
    os.environ['LANG'] = old_lang


def _extract_subs(srcpath, trackno):
    """Extract the subs from a file."""
    _, temppath = tempfile.mkstemp()
    if srcpath.endswith('.mp4'):
        cmd = ['ffmpeg', '-i', srcpath, '-map', '0:{}'.format(trackno), temppath]
    elif srcpath.endswith('.mkv'):
        cmd = ['mkvextract', 'tracks', srcpath, '{}:{}'.format(trackno, temppath)]
    else:
        raise ValueError("Not supported format to extract subs: {}".format(srcpath))

    subprocess.run(cmd, check=True)

    with open(temppath, 'rt', encoding='utf8') as fh:
        content = fh.read()
    os.unlink(temppath)
    return content


def _get_subs(filepath):
    """Get the subtitle from the file."""
    # find out in which track it is
    cmd = ["mkvmerge", "-i", filepath]
    with lang_changer('C'):
        cp = subprocess.run(cmd, stdout=subprocess.PIPE, check=True, encoding='utf8')
    lines = cp.stdout.split("\n")

    tracks = []
    for line in lines:
        m = RE_TRACK.match(line)
        if m:
            info = m.groupdict()
            if info['kind'] == 'subtitles':
                tracks.append((info['trackid'], info['spec'].lower()))

    for track_id, spec in tracks:
        if 'srt' in spec:
            subtype = 'srt'
            break
        if 'substationalpha' in spec:
            subtype = 'ssa'
            break
        print("Warning: unknown sub type {!r} (track {})".format(spec, track_id))
    else:
        # nothing valid found
        return

    # get the file content for given track
    content = _extract_subs(filepath, track_id)

    # process that content according to its type
    processing_func = SUB_LOADERS[subtype]
    return processing_func(content)


class Subtitles:
    def __init__(self, subs):
        self.subs = subs
        if subs is None:
            self.exists = False
            print("Subtitles not found")
        else:
            self.exists = True
            print("Subtitles ok")

    @classmethod
    def from_movie(cls, srcpath):
        subs = _get_subs(srcpath)
        return cls(subs)

    @classmethod
    def from_subs(cls, srcpath):
        with open(srcpath, 'rt', encoding='utf8') as fh:
            content = fh.read()

        # process that content according to its type
        loader = SUB_LOADERS[srcpath[-3:]]
        subs = loader(content)
        return cls(subs)

    def cut(self, time_from, time_to, filepath):
        """Cut the sub at those times and save it."""
        # get the first sub inside the desired range
        subs = iter(self.subs)
        while True:
            try:
                item = next(subs)
            except StopIteration:
                # the whole file ended without finding the start of this cut's sub
                return

            if item.tfrom > time_to:
                # this sub item belongs already to next cut, there was nothing for *this* cut!
                return

            if item.tfrom > time_from:
                break

        # get all the rest until we exit the range
        all_items = [item]
        while True:
            try:
                item = next(subs)
            except StopIteration:
                break

            if item.tfrom > time_to:
                break
            all_items.append(item)

        # rescale all the items to the new beginning
        newitems = []
        for item in all_items:
            newfrom = item.tfrom - time_from
            newto = item.tto - time_from
            newitems.append(subtitles.SubItem(newfrom, newto, item.text))

        # save all items to disk
        subtitles.save_srt(newitems, filepath)
        return True


def main(srcpath, cuts, subsfilepath):
    print("Processing", repr(srcpath))
    master_clip = VideoFileClip(srcpath)
    if subsfilepath:
        master_subs = Subtitles.from_subs(subsfilepath)
    else:
        master_subs = Subtitles.from_movie(srcpath)

    for tf, tt, dest in cuts:
        if dest == '-':
            print("Skipping from {!r} to {!r}".format(tf, tt))
            continue

        _, temp_video = tempfile.mkstemp(dir='.', suffix='.mp4')
        print("Cutting: {!r} - {!r}".format(tf, tt))
        video_clip = master_clip.subclip(tf, tt)
        video_clip.write_videofile(temp_video)

        if not master_subs.exists:
            # we're done
            dest += '.mp4'
            os.rename(temp_video, dest)
            print("Done: {!r}".format(dest))
            continue

        print("Preparing subtitles")
        _, temp_subs = tempfile.mkstemp(dir='.', suffix='.srt')
        subs_ok = master_subs.cut(tf, tt, temp_subs)
        if not subs_ok:
            # there wasn't any sub for this clip, really
            dest += '.mp4'
            os.rename(temp_video, dest)
            print("Done: {!r}".format(dest))
            continue

        print("Merging")
        _, temp_merged = tempfile.mkstemp(dir='.', suffix='.mkv')
        cmd = ['merge_video_subs.py', '--name={}'.format(temp_merged), temp_video, temp_subs]
        subprocess.run(cmd, check=True)
        os.remove(temp_video)
        os.remove(temp_subs)

        dest += '.mkv'
        os.rename(temp_merged, dest)
        print("Done: {!r}".format(dest))


def _parse(infolines):
    """Parse the info lines."""
    cuts = []
    for item in infolines:
        print("Parsing", repr(item))
        parts = item.split(maxsplit=2)
        if len(parts) != 3:
            print("Bad parts (must be 3):", parts)
            exit()

        tf, tt, dest = parts

        try:
            tf = subtitles.time_sub2stamp(tf)
            tt = subtitles.time_sub2stamp(tt)
        except ValueError:
            print("Bad time info:", parts)
            exit()

        result = (tf, tt, dest.strip())
        print("    got", result)
        cuts.append(result)
    return cuts


USAGE = """
USAGE:
    videocutter.py srcvideo time_from time_to destvideo
    videocutter.py srcvideo cutinfo

        optionally:  --sub subsfilepath

times can be decimals, or NN:NN:NN.NN format.

"cutinfo" is a file with several lines like:
    time_from time_to destvideo
"""

if __name__ == "__main__":
    if '--sub' in sys.argv:
        idx = sys.argv.index('--sub')
        del sys.argv[idx]
        subs_filepath = sys.argv.pop(idx)
    else:
        subs_filepath = None

    if len(sys.argv) == 3:
        with open(sys.argv[2], 'rt', encoding='utf8') as fh:
            cut_raw = fh.readlines()
    elif len(sys.argv) == 5:
        cut_raw = [" ".join(sys.argv[2:5])]
    else:
        print(USAGE)
        exit()

    cut_info = _parse(cut_raw)
    main(sys.argv[1], cut_info, subs_filepath)
